/***************************************************************************
                          komportview.cpp  -  Komport Serial Port Communicator
                             -------------------
    begin                : Mon Feb 17 00:05:54 EST 2003
    copyright            : (C) 2003 by Mike Sharkey
    email                : michael@sharkey.servebeer.com                              1
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qprinter.h>
#include <qpainter.h>
#include <qrect.h>
#include <qfont.h>
#include <qfontmetrics.h>

// application specific includes
#include "komportview.h"
#include "komportdoc.h"
#include "komport.h"

#define min(x,y) (x<y?x:y)
#define max(x,y) (x<y?y:x)

#define inherited QScrollView

#define _DEFAULT_FOREGROUND_   QApplication::palette().active().foreground()
#define _DEFAULT_BACKGROUND_   QApplication::palette().active().background()

KomportView::KomportView(QWidget *parent, const char *name)
: inherited(parent, name)
, mBlink(false)
, mBold(false)
, mReverse(false)
, mUnderline(false)
, mNotify(true)
, mCursorOn(true)
, mCursorState(false)
, mBlinkState(false)
, mInSelection(false)
, mHasSelection(false)
, mInTimerScroll(false)
{  
  mCells.setAutoDelete(true);
  initSettings();
  viewport()->setBackgroundMode(PaletteBase);
  mForegroundColor = viewport()->palette().active().foreground();
  mBackgroundColor = viewport()->palette().active().background();
  
  QObject::connect(QApplication::clipboard(),SIGNAL(selectionChanged()),this,SLOT(slotSelectionChanged()));

  mCursorTimer = startTimer( 500 );
  mBlinkTimer = startTimer( 1000 );
  mAutoScrollTimer = startTimer( 250 );
  mStatusTimer = startTimer( 100 );
  
  setCellSize();
  setEnabled(true);
  setFocusPolicy(StrongFocus);

  mEmulation = new KomportEmulation( getSerial(), this );
  QObject::connect(this,SIGNAL(keyPressed(QKeyEvent*)),mEmulation,SLOT(slotKeyPressed(QKeyEvent*)));
  QObject::connect(this,SIGNAL(simKeyPressed(QChar)),mEmulation,SLOT(slotSimKeyPressed(QChar)));
  QObject::connect(mEmulation,SIGNAL(cursorOn()),this,SLOT(slotCursorOn()));
  QObject::connect(mEmulation,SIGNAL(cursorOff()),this,SLOT(slotCursorOff()));
  
  
  QObject::connect(getSerial(),SIGNAL(receivedChar(char)),this,SLOT(slotReceivedChar(char)));
  QObject::connect(getSerial() ,SIGNAL(queueOverflow()),this,SLOT(slotQueueOverflow()));
  QObject::connect(this,SIGNAL(updateStatus(const QString &)),(KomportApp*)parent,SLOT(slotStatusMsg(const QString &)));
}

KomportView::~KomportView()
{
}

KomportDoc *KomportView::getDocument() const
{
  KomportApp *theApp=(KomportApp *) parentWidget();

  return theApp->getDocument();
}

void KomportView::print(QPrinter *pPrinter)
{
  for ( int copies=0; copies < pPrinter->numCopies(); copies++ ) {
      QPainter paint;
      paint.begin(pPrinter);
      QRect bounds = paint.viewport();
      QString str="";
      if ( hasSelection() ) {
          // draw the contents of the mouse clipboard...
          str = QApplication::clipboard()->text( QClipboard::Clipboard );
      } else {
          // draw the contents of the screen...
          QRect rect = contentCellsRect();
          for ( int y=rect.y(); y < rect.height(); y++ ) {
            for ( int x=rect.x(); x < rect.width(); x++ ) {
                  KomportCell* cell = getCell(x,y);
                  if ( cell ) {
                    str += cell->character();
                  }
              }
              str += QChar( '\n' );
          }
      }
      paint.drawText( bounds, AlignAuto|PlainText, str );
      // eject the page...
      paint.end();
  }
}

/** draw a cell */
void KomportView::paintCell( QPainter* _paint, KomportCell* cell, QRect _bounds, bool _cursorOn ) {
  if ( cell != NULL ) {
    QString str( cell->character() );
    QColor fillColor;
    QColor textColor;
    QColor cellBackground = cell->backgroundColor();
    QColor cellForeground = cell->foregroundColor();

    // determine forground and background color...
    if ( cell->select() ) {
        fillColor =  QApplication::palette().active().highlight();
        textColor = QApplication::palette().active().highlightedText();
    } else  if ( _cursorOn && hasFocus() ) {
      fillColor = cell->reverse() ? cellBackground: cellForeground;
      textColor = cell->reverse() ? cellForeground:cellBackground;
    } else {
      fillColor = cell->reverse() ? cellForeground : cellBackground;
      textColor = cell->reverse() ? cellBackground : cellForeground;
    }

    // do other attributes and draw the character...
    _paint->fillRect( _bounds, QBrush( fillColor ) );
    _paint->setPen( textColor );
    QFont font = _paint->font();
    font.setBold( cell->bold() );
    font.setUnderline(cell->underline());
    _paint->setFont(font);
    if ( ( !cell->blink() || (cell->blink() &&  !mBlinkState) ) )
      _paint->drawText ( _bounds, Qt::AlignCenter, str );
    if ( _cursorOn && !hasFocus() ) { // out-of-focus cursor box
      _paint->drawRect( _bounds );
    }
  }
}

/** The function should draw the rectangle (clipx, clipy, clipw, cliph) of the contents using painter p.
The clip rectangle is in the scrollview's coordinates.
The clip rectangle and translation of the painter p is already set appropriately. */
void KomportView::drawContents ( QPainter* _paint, int clipx, int clipy, int clipw, int cliph )
{
  QRect rect = QRect( clipx, clipy, clipw, cliph );
  QPoint topLeft = cellPoint( rect.topLeft() );
  QPoint bottomRight = cellPoint( rect.bottomRight() );
  for( int y=topLeft.y(); y <= bottomRight.y(); y++ ) {
    for( int x=topLeft.x(); x <= bottomRight.x(); x++ ) {
      updateCell( _paint, x,y );
    }
  }
}

/** re-draw a cell */
void KomportView::updateCell(QPainter* _paint, int _x,int _y){
  int w = cellWidth();
  int h = cellHeight();
  QRect cellRect( _x*w, _y*h, w, h );
  paintCell( _paint, getCell(_x, _y) , cellRect, cursorCellOn( _x, _y)   );
}

/** update a cell */
void KomportView::updateCell(QPainter* _paint, QPoint _p){
  updateCell( _paint, _p.x(), _p.y() );
}

/** re-draw a cell */
void KomportView::updateCell(int _x,int _y){
  QPoint contentsPt( _x*cellWidth(), _y*cellHeight() );
  QPoint viewportPt = contentsToViewport( contentsPt );
  QRect cellRect( viewportPt.x(), viewportPt.y(), cellWidth(), cellHeight() );
  QPainter paint( viewport() );
  paintCell( &paint, getCell(_x, _y) , cellRect, cursorCellOn( _x, _y)   );
  paint.end();
}

/** update a cell */
void KomportView::updateCell(QPoint _p){
  updateCell( _p.x(), _p.y() );
}

/** This event handler is called whenever the QScrollView receives a mousePressEvent(): the press position in e is translated to be a point on the contents. */
void KomportView::contentsMousePressEvent( QMouseEvent* _e )
{
  if ( _e->button() == Qt::LeftButton ) {
    mAnchor = cellPoint( _e->pos() );
    select( mAnchor, mAnchor );
    mInSelection = true;
  }
}

/**  This event handler is called whenever the QScrollView receives a mousePressEvent(): the press position in e is translated to be a point on the contents. */
void KomportView::contentsMouseReleaseEvent( QMouseEvent* _e )
{
  if( _e->button() == Qt::LeftButton && mInSelection ) {
    select( mSelectRect.topLeft(), mSelectRect.bottomRight(), true );
    mInSelection = false;
  }
}

/** This event handler is called whenever the QScrollView receives a mouseMoveEvent(): the mouse position in e is translated to be a point on the contents. */
void KomportView::contentsMouseMoveEvent( QMouseEvent* _e )
{
  if ( mInSelection ) {
    mMousePos = cellPoint( _e->pos() );
    select( mAnchor, mMousePos );
    ensureVisible( _e->pos().x(), _e->pos().y() );
  }
}

/** key press event */
void KomportView::keyPressEvent(QKeyEvent* _e){
  resetScroll();
  emit keyPressed(_e);
}

/** key release event */
void KomportView::keyReleaseEvent(QKeyEvent* _e){
}

/** timer event */
void KomportView::timerEvent(QTimerEvent* _e) {
  if ( _e->timerId() == mBlinkTimer ) {
    mBlinkState = !mBlinkState;
    QRect rect = contentCellsRect();
    for ( int x=rect.x(); x < rect.width(); x++ ) {
      for ( int y=rect.y(); y < rect.height(); y++ ) {
        KomportCell* cell = getCell(x,y);
        if ( cell && cell->blink() ) {
          updateCell(x,y);
        }
      }
    }
  } else if ( _e->timerId() == mCursorTimer ) {
    mCursorState = !mCursorState;
    updateCell(cursor().x(),cursor().y());
  } else if (_e->timerId() == mAutoScrollTimer ) {
    //debug( "auto-scroll" );
  }  else if (_e->timerId() == mStatusTimer ) {
      if ( getSerial()->level() > 0.0 ) {
          QString str = "Ready.  RX(" + QString::number(getSerial()->level(),'g',2) + "%)" ;
          emit updateStatus(str);
      }
  }
}

/** notify cursor position has changde */
void KomportView::cursorChanged(QPoint _old, QPoint _new)
{
  updateCell( _old );
  mCursorState=true;
  updateCell( _new );
}

/** update a row */
void KomportView::updateRow(int _row){
  for( int x=0; x < arrayWidth(); x++ )
    updateCell(x,_row);
}


/** received a char */
void KomportView::slotReceivedChar(char _ch){
  resetScroll();
}

/** get serial port */
KomportSerial* KomportView::getSerial(){
  return getDocument()->getSerial();
}

/** make a selection by cell coordinates */
void KomportView::select(QPoint start, QPoint end, bool clip){
    QRect newRect = QRect( start, end );
    deselect( newRect.normalize() );
    mSelectRect = newRect.normalize();
    if ( start != end ) {
      int startX = min(start.x(),end.x());
      int startY = min(start.y(),end.y());
      int endX = max(start.x(),end.x());
      int endY = max(start.y(),end.y());
      KomportCell* cell=NULL;
      QString str;
      for( int y=startY; y <= endY; y++ ) {
          for( int x=startX; x <= endX; x++ ) {
              cell = getCell(x,y);
              if ( cell != NULL ) {
                  getCell(x,y)->setSelect( true );;
                  updateCell(x,y);
                  if ( clip )
                      str += cell->character();
              }
          }
          if ( clip && cell != NULL ) {
              str += QChar('\n');
          }
      }
      if ( clip ) {
          QApplication::clipboard()->setText(str,QClipboard::Selection);
          mHasSelection = true;
          emit viewModified(this);
      }
    }
}

/** clear selection */
void KomportView::deselect(){
    QPoint start = mSelectRect.topLeft();
    QPoint end = mSelectRect.bottomRight();
    int startX = min(start.x(),end.x());
    int startY = min(start.y(),end.y());
    int endX = max(start.x(),end.x());
    int endY = max(start.y(),end.y());
    //debug( "deselect\t( %d, %d )\t(%d, %d)", startX, startY, endX, endY );
    for( int y=startY; y <= endY; y++ ) {
        for( int x=startX; x <= endX; x++ ) {
            KomportCell* cell = getCell(x,y);
            if ( cell != NULL ) {
                cell->setSelect(false);
                updateCell(x,y);
            }
        }
    }
    mSelectRect = QRect( QPoint(0,0), QPoint(0,0) );
    mHasSelection = false;
    emit viewModified( this );
}

/** clear selection */
void KomportView::deselect(QRect _bounds){
  QPoint start = mSelectRect.topLeft();
  QPoint end = mSelectRect.bottomRight();
  int startX = min(start.x(),end.x());
  int startY = min(start.y(),end.y());
  int endX = max(start.x(),end.x());
  int endY = max(start.y(),end.y());
  for( int y=startY; y <= endY; y++ ) {
    for( int x=startX; x <= endX; x++ ) {
      if ( !_bounds.contains( x, y ) ) {
        KomportCell* cell = getCell(x,y);
        if ( cell != NULL ) {
          cell->setSelect(false);
          updateCell(x,y);
        }
      }
    }
  }
  mSelectRect = QRect( QPoint(0,0), QPoint(0,0) );
  mHasSelection = false;
  emit viewModified( this );
}

/** No descriptions */
void KomportView::slotSelectionChanged() {
    if ( !mInSelection ){
        deselect();
    }
    emit viewModified(this);
}

/** has a text selection */
bool KomportView::hasSelection(){
    return mHasSelection;
}

/** simulated key press for inserting from clipboard, etc... */
void KomportView::slotSimKeyPressed(QChar _c){
  resetScroll();
  emit simKeyPressed(_c);
}

/** reset scrollbar */
void KomportView::resetScroll(){
  verticalScrollBar()->setValue( verticalScrollBar()->maxValue() );
}

/** set the number of lines in the scroll buffer */
void KomportView::setScrollBuffer(int _sz){
  debug( "setScrollBuffer %d lines", _sz );
  resetScroll();
}
/** turn cursor on (visible) */
void KomportView::slotCursorOn(){
    mCursorOn=true;
}
/** turn cursor off (invisible) */
void KomportView::slotCursorOff(){
    mCursorOn=false;
}
/** rx queue has overflowed */
void KomportView::slotQueueOverflow(){
    QString str = "Ready. RX (OVERFLOW)" ;
    emit updateStatus(str);
}
/** convert a local pixel coord to a cell coord */
QPoint KomportView::cellPoint(QPoint _pt){
  QPoint cp = QPoint( _pt.x()  / cellWidth(), _pt.y() / cellHeight() );
  //debug( "pix=%d,%d cell=%d,%d", _pt.x(), _pt.y(), cp.x(), cp.y() );
  return cp;
}

/** returns enclosing rectangle of visible cells */
QRect KomportView::contentCellsRect()
{
  QPoint topLeft = cellPoint( QPoint( contentsX(), contentsY() ) );
  QPoint bottomRight = cellPoint( QPoint( contentsX()+contentsWidth(), contentsY()+contentsHeight() ) );
  return QRect( topLeft, bottomRight );
}

/** Returns true if the specified cell is the cursor cell, and the current cursor state is true, and cursor is enabled.
 */
bool KomportView::cursorCellOn(int _x, int _y){
    return ( mCursorOn && (cursor() == QPoint( _x, _y ) && mCursorState) );
}

 /** initialize default settings */
void KomportView::initSettings(){
  setCellSize();
  setArraySize( QSize( 80, 25 ) );
  setCursor( QPoint( 0, 0 ) );
}

/** set the cell array size .
 */
void KomportView::setArraySize(QSize _sz){
  mArraySize = _sz;
  int curcnt = mCells.count();
  int newcnt = _sz.width()*_sz.height();
  int index;
  if ( curcnt==0 ) {
      for( index=0; index < mArraySize.width()*mArraySize.height(); index++ ) {
          KomportCell *pCell = new KomportCell();
          mCells.append(pCell);
      }
  } else  if ( curcnt > newcnt ) {
      int diff = curcnt - newcnt;
      for( index=0; index < diff; index++ )
          mCells.removeFirst();
  } else if ( curcnt < newcnt  ) {
      int diff = newcnt - curcnt;
      for( index=0; index < diff; index++ )  {
          KomportCell *pCell = new KomportCell();
          mCells.append(pCell);
      }
  }
  int w = arrayWidth()*cellWidth();
  int h = arrayHeight()*cellHeight();
  resizeContents( w, h );
}

/** calculates cell size based on font. */
void KomportView::setCellSize(){
  QFontMetrics fm = fontMetrics();
  setCellSize(QSize(fm.width(QChar('H')),fm.height()));
  QSize sz(width(),height());
  setMinimumSize(sz);
  setMaximumSize(sz);
}

/** set cell size */
void KomportView::setCellSize(QSize _sz){
  mCellSize=_sz;
}

/** return calculated dimensions in pixels based on cell size and array size */
QSize KomportView::size(){
  QScrollBar* sb = verticalScrollBar();
  return QSize( (arrayWidth()*cellWidth())+(sb->isVisible()?sb->width():0), arrayHeight()*cellHeight() );
}

/** get array size */
QSize KomportView::arraySize(){
  return mArraySize;
}

/** get cell size */
QSize KomportView::cellSize(){
  return mCellSize;
}

/** return calculated width based on cell width * array width */
int KomportView::width(){
  return size().width();
}

/** return calculated height based on cell height * array height */
int KomportView::height(){
  return size().height();
}

/** array width  */
int KomportView::arrayWidth(){
  return arraySize().width();
}

/** array height */
int KomportView::arrayHeight(){
  return arraySize().height();
}

/** cell width */
int KomportView::cellWidth(){
  return cellSize().width();
}

/** cell height */
int KomportView::cellHeight(){
  return cellSize().height();
}

/** get a pointer to the cell from location (x,y) */
KomportCell* KomportView::getCell(int _x,int _y){
  unsigned int index = (arrayWidth()*_y)+_x;
  if ( index < mCells.count() ) {
    return mCells.at( index );
  } else {
    return NULL;
  }
}

/** get a pointer to the cell from location (x,y) */
KomportCell* KomportView::getCell(QPoint _p){
  return getCell(_p.x(),_p.y());
}

/** draw a character into a cell */
void KomportView::drawChar(QChar _c, int _x, int _y){
  drawChar(_c,QPoint(_x,_y));
}

/** draw a character into a cell */
void KomportView::drawChar(QChar _c, QPoint _p){
  QChar old = getChar(_p);
  if (old != _c) {
    getCell(_p)->setCharacter(_c);
    setCellAttributes(_p);
    if (mNotify) updateCell(_p);
  }
}

/** cursor position */
QPoint KomportView::cursor(){
  return mCursor;
}

/** cursor position */
void KomportView::setCursor(QPoint _p){
  QPoint old = mCursor;
  if (old != _p) {
    mCursor = _p;
    if (mNotify) cursorChanged(old,_p);
  }
}

/** Insert a character at the current cursor position and advance cursor. Scroll if advance is passed the last column. */
void KomportView::putChar(QChar _ch){
  drawChar(_ch,cursor());
  advanceCursor();
}

/** move the cursor once cell forward, scolling or wrapping as nessesary */
void KomportView::advanceCursor(){
  QPoint pos = cursor();
  pos.setX(pos.x()+1);
  if ( pos.x() >= arrayWidth() ) {
    pos.setX(0);
    pos.setY(pos.y()+1);
    setCursor(pos);
    if ( pos.y() >= arrayHeight() ) {
      pos.setY(arrayHeight()-1);
      setCursor(pos);
      scrollUp();
    }
  } else {
    setCursor(pos);
  }
}

/** retrieve a character from a cell */
QChar KomportView::getChar(QPoint _p){
  return getCell(_p)->character();
}

/** scroll screen up one line */
void KomportView::scrollUp(){
    setArraySize( QSize( arrayWidth(), arrayHeight()+1 ) ); // append a new row
    clearRow(arrayHeight()-1);
    QPoint pos = cursor();
    pos.setY(arrayHeight()-1);
    setCursor(pos);
    resetScroll();
}

/** copy a row  */
void KomportView::copyRow(int _dst, int _src){
  int w = arrayWidth();
  for( int x=0; x < w; x++ ) {
     KomportCell* srcCell = getCell(x,_src);
     KomportCell* dstCell = getCell(x,_dst);
     dstCell->copy( srcCell );
  }
  if ( mNotify ) updateRow(_dst);
}

/** clear a row */
void KomportView::clearRow(int _row){
  int w = arrayWidth();
  for( int x=0; x < w; x++ ) {
    getCell(x,_row)->clear();
  }
  if ( mNotify ) updateRow(_row);
}

/** update all cells */
void KomportView::updateAll(){
  //debug( "updateAll" );
  QRect bounds = contentCellsRect(); // get current viewport rect in cell coords.
  for( int y=bounds.y(); y < arrayHeight(); y++ )
    updateRow(y);
}

/** clear all cells */
void KomportView::clear(){
  QRect bounds = contentCellsRect(); // get current viewport rect in cell coords.
  for( int y=bounds.y(); y < arrayHeight(); y++ )
    clearRow(y);
}

/** clear a cell */
void KomportView::clear(QPoint _p)
{
  getCell(_p)->clear();
  updateCell(_p);
}


/** clear a cell */
void KomportView::clear(int _x,int _y)
{
  clear(QPoint(_x,_y));
}

/** set cell attributes */
void KomportView::setCellAttributes(QPoint _p){
  KomportCell* c = getCell(_p);
  c->setBlink(blink());
  c->setBold(bold());
  c->setReverse(reverse());
  c->setUnderline(underline());
  c->setForegroundColor(foregroundColor());
  c->setBackgroundColor(backgroundColor());
}

/** default foreground color */
QColor KomportView::defaultForegroundColor(){
  return _DEFAULT_FOREGROUND_;
}

/** default background color */
QColor KomportView::defaultBackgroundColor(){
  return _DEFAULT_BACKGROUND_;
}

/** clear to end of line */
void KomportView::clearEOL(){
  for( int x=cursor().x(); x < arrayWidth(); x++ ) {
    clear(x,cursor().y());
  }
}
/** set a cell */
void KomportView::setCell( QPoint _p, KomportCell* _cell){
    getCell(_p)->copy(_cell);
}

